# lingo - payment and billing system
# Copyright (C) 2022  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime

import pytz
from django.http import Http404, HttpResponse
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_noop as N_
from rest_framework import permissions
from rest_framework.views import APIView
from weasyprint import HTML

from lingo.agendas.models import Agenda
from lingo.api import serializers
from lingo.api.utils import APIErrorBadRequest, Response
from lingo.invoicing.models import DraftInvoice, DraftInvoiceLine, InjectedLine, Invoice, Payment, Regie
from lingo.pricing.models import AgendaPricing, PayerError


class AgendaCheckTypeList(APIView):
    permission_classes = ()

    def get(self, request, agenda_identifier=None, format=None):
        agenda = get_object_or_404(Agenda, slug=agenda_identifier)

        check_types = []
        if agenda.check_type_group:
            check_types = [
                {'id': x.slug, 'text': x.label, 'kind': x.kind}
                for x in agenda.check_type_group.check_types.filter(disabled=False)
            ]
        return Response({'data': check_types})


agenda_check_type_list = AgendaCheckTypeList.as_view()


class PricingCompute(APIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.PricingComputeSerializer

    def get(self, request, format=None):
        serializer = self.serializer_class(data=request.query_params)
        if not serializer.is_valid():
            raise APIErrorBadRequest(N_('invalid payload'), errors=serializer.errors)

        return Response({'data': serializer.compute(self.request)})


pricing_compute = PricingCompute.as_view()


class InvoicingRegies(APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, format=None):
        return Response(
            {
                'data': [
                    {'id': regie.slug, 'text': regie.label, 'slug': regie.slug}
                    for regie in Regie.objects.order_by('label')
                ]
            }
        )


invoicing_regies = InvoicingRegies.as_view()


class InvoiceMixin:
    def get_payer_external_ids(self, request, regie, nameid):
        if not nameid:
            raise Http404
        payer_external_ids = set()
        agenda_pricing_qs = (
            AgendaPricing.objects.filter(
                agendas__regie=regie,
            )
            .distinct()
            .order_by('pk')
        )
        for agenda_pricing in agenda_pricing_qs:
            try:
                payer_external_ids.add(agenda_pricing.get_payer_external_id_from_nameid(request, nameid))
            except PayerError:
                pass
        return list(payer_external_ids)


class InvoicingInvoices(InvoiceMixin, APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get_invoices_queryset(self, request, regie):
        payer_external_ids = self.get_payer_external_ids(request, regie, request.GET.get('NameID'))
        return Invoice.objects.filter(
            regie=regie,
            remaining_amount__gt=0,
            date_publication__lte=datetime.date.today(),
            payer_external_id__in=payer_external_ids,
        ).order_by('-created_at')

    def get(self, request, regie_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        invoices = self.get_invoices_queryset(request, regie)
        return Response({'data': [invoice.normalize() for invoice in invoices]})


invoicing_invoices = InvoicingInvoices.as_view()


class InvoicingHistoryInvoices(InvoicingInvoices):
    def get_invoices_queryset(self, request, regie):
        payer_external_ids = self.get_payer_external_ids(request, regie, request.GET.get('NameID'))
        return Invoice.objects.filter(
            regie=regie,
            remaining_amount=0,
            date_publication__lte=datetime.date.today(),
            payer_external_id__in=payer_external_ids,
        ).order_by('-created_at')


invoicing_history_invoices = InvoicingHistoryInvoices.as_view()


class InvoicingInvoice(InvoiceMixin, APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, regie_identifier, invoice_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        payer_external_ids = self.get_payer_external_ids(request, regie, request.GET.get('NameID'))
        invoice = get_object_or_404(
            Invoice,
            uuid=invoice_identifier,
            regie=regie,
            date_publication__lte=datetime.date.today(),
            payer_external_id__in=payer_external_ids,
        )
        return Response({'data': invoice.normalize()})


invoicing_invoice = InvoicingInvoice.as_view()


class InvoicingInvoicePDF(InvoiceMixin, APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def get(self, request, regie_identifier, invoice_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        payer_external_ids = self.get_payer_external_ids(request, regie, request.GET.get('NameID'))
        invoice = get_object_or_404(
            Invoice,
            uuid=invoice_identifier,
            regie=regie,
            date_publication__lte=datetime.date.today(),
            payer_external_id__in=payer_external_ids,
        )
        result = invoice.html()
        html = HTML(string=result)
        pdf = html.write_pdf()
        response = HttpResponse(pdf, content_type='application/pdf')
        response['Content-Disposition'] = 'attachment; filename="%s.pdf"' % invoice.formatted_number
        return response


invoicing_invoice_pdf = InvoicingInvoicePDF.as_view()


class InvoicingInvoicePay(APIView):
    # XXX ?
    authentication_classes = []
    permission_classes = ()

    def post(self, request, regie_identifier, invoice_identifier):
        data = request.data
        order_id = data.get('transaction_id')
        order_date = data.get('transaction_date')
        if order_date:
            try:
                order_date = pytz.utc.localize(datetime.datetime.strptime(order_date, '%Y-%m-%dT%H:%M:%S'))
            except ValueError:
                order_date = None
        invoice = get_object_or_404(
            Invoice,
            uuid=invoice_identifier,
            regie__slug=regie_identifier,
            remaining_amount__gt=0,
        )
        Payment.make_payment(
            regie=invoice.regie,
            invoices=[invoice],
            amount=invoice.remaining_amount,
            payment_type='online',
            order_id=order_id,
            order_date=order_date,
        )
        return Response({'data': True})


invoicing_invoice_pay = InvoicingInvoicePay.as_view()


class InvoicingDraftInvoices(APIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.DraftInvoiceSerializer

    def post(self, request, regie_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        serializer = self.serializer_class(data=request.data)
        if not serializer.is_valid():
            raise APIErrorBadRequest(N_('invalid payload'), errors=serializer.errors)

        invoice = DraftInvoice.objects.create(regie=regie, **serializer.validated_data)

        return Response({'data': {'draft_invoice_id': str(invoice.uuid)}})


invoicing_draft_invoices = InvoicingDraftInvoices.as_view()


class InvoicingDraftInvoiceLines(APIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.DraftInvoiceLineSerializer

    def post(self, request, regie_identifier, invoice_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        invoice = get_object_or_404(DraftInvoice, regie=regie, uuid=invoice_identifier, pool__isnull=True)
        serializer = self.serializer_class(data=request.data)
        if not serializer.is_valid():
            raise APIErrorBadRequest(N_('invalid payload'), errors=serializer.errors)

        line = DraftInvoiceLine.objects.create(
            invoice=invoice,
            payer_external_id=invoice.payer_external_id,
            payer_first_name=invoice.payer_first_name,
            payer_last_name=invoice.payer_last_name,
            payer_demat=invoice.payer_demat,
            payer_direct_debit=invoice.payer_direct_debit,
            status='success',
            **serializer.validated_data,
        )

        return Response({'data': {'draft_line_id': line.pk}})


invoicing_draft_invoice_lines = InvoicingDraftInvoiceLines.as_view()


class InvoicingDraftInvoiceClose(APIView):
    permission_classes = (permissions.IsAuthenticated,)

    def post(self, request, regie_identifier, invoice_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        invoice = get_object_or_404(DraftInvoice, regie=regie, uuid=invoice_identifier, pool__isnull=True)
        final_invoice = invoice.promote()

        return Response({'data': {'invoice_id': str(final_invoice.uuid)}})


invoicing_draft_invoice_close = InvoicingDraftInvoiceClose.as_view()


class InjectedLines(APIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.InjectedLineSerializer

    def post(self, request, regie_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        serializer = self.serializer_class(data=request.data)
        if not serializer.is_valid():
            raise APIErrorBadRequest(N_('invalid payload'), errors=serializer.errors)

        instance = InjectedLine.objects.create(regie=regie, **serializer.validated_data)
        return Response({'err': 0, 'id': instance.pk})


injected_lines = InjectedLines.as_view()


class Payments(APIView):
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.PaymentSerializer

    def post(self, request, regie_identifier):
        regie = get_object_or_404(Regie, slug=regie_identifier)
        serializer = self.serializer_class(data=request.data, regie=regie)
        if not serializer.is_valid():
            raise APIErrorBadRequest(N_('invalid payload'), errors=serializer.errors)

        instance = Payment.make_payment(
            regie=regie,
            invoices=serializer._invoices,
            amount=serializer.validated_data['amount'],
            payment_type=serializer.validated_data['payment_type'],
        )
        return Response({'err': 0, 'id': instance.pk})


payments = Payments.as_view()
